import os
import stat


def chmod_executable(filename):
    mode = os.stat(filename).st_mode
    new_mode = mode | stat.S_IXUSR | stat.S_IXGRP | stat.S_IXOTH
    if new_mode != mode:
        os.chmod(filename, new_mode)



from ftw.recipe.deployment.utils import chmod_executable
import os
import os.path


def create_pack_script(recipe):
    """Creates a pack script for packing all storages.
    """
    # Packing needs a zeo server part
    if not recipe.zeo_parts:
        return []

    storages = []

    # Determine storages from zeo server parts
    for zeo_part in recipe.zeo_parts:
        file_storage = recipe.buildout[zeo_part].get('storage-number', '1')
        blob_storage = None
        if recipe.buildout[zeo_part].get('blob-storage', 'blobstorage'):
            blob_storage = recipe.buildout[zeo_part].get(
                'blob-storage', 'blobstorage')
            blob_storage = os.path.join(recipe.buildout_dir, 'var', blob_storage)
        storages.append((file_storage, blob_storage))

    # Determine storages from c.r.filestorage parts
    for fs_part in recipe.filestorage_parts:
        for part in recipe.buildout[fs_part].get('parts', '').split():
            file_storage = subpart_option(
                recipe.buildout, fs_part, part, 'zeo-storage',
                default='%(fs_part_name)s')
            blob_storage = None
            if subpart_option(recipe.buildout, fs_part, part, 'blob-storage'):
                blob_storage = subpart_option(
                    recipe.buildout, fs_part, part, 'blob-storage',
                    default=os.path.join('var', 'blobstorage-%(fs_part_name)s'))
                if not blob_storage.startswith(os.path.sep):
                    blob_storage = os.path.join(
                        recipe.buildout['buildout']['directory'], blob_storage)
            storages.append((file_storage, blob_storage))

    # Create script
    created_files = []
    script = "#!/bin/sh\n"
    zeopack = os.path.join(recipe.buildout_dir, 'bin', 'zeopack')
    for storage in storages:
        if storage[1]:
            script += "%(zeopack)s -S %(file_storage)s -B %(blob_storage)s" % dict(
                zeopack=zeopack, file_storage=storage[0], blob_storage=storage[1])
            script += build_packed_log_entry_command(recipe, storage[0], storage[1])
            script += '\n'
        else:
            script += "%(zeopack)s -S %(file_storage)s" % dict(
                zeopack=zeopack, file_storage=storage[0])
            script += build_packed_log_entry_command(recipe, storage[0])
            script += '\n'
    script_path = os.path.join(recipe.buildout_dir, 'bin', 'packall')
    script_file = open(script_path, 'w')
    script_file.write(script)
    script_file.close()
    chmod_executable(script_path)
    created_files.append(script_path)

    # Create symlink
    symlink_dir = recipe.options.get('packall-symlink-directory', None)
    if symlink_dir:
        # Try to create the symlink directory
        if not os.path.isdir(symlink_dir):
            try:
                os.makedirs(symlink_dir)
            except OSError:
                pass
        if os.path.isdir(symlink_dir):
            link_path = os.path.join(symlink_dir, recipe.buildout_name)
            if os.path.exists(link_path):
                os.remove(link_path)
            os.symlink(script_path, link_path)
            created_files.append(link_path)
    return created_files


def build_packed_log_entry_command(recipe, storage, blobstorage=None):
    if storage == '1':
        storage = 'Data'
    if blobstorage:
        blobstorage = os.path.basename(blobstorage)

    if blobstorage:
        message = 'packed {} ({})'.format(storage, blobstorage)
    else:
        message = 'packed {}'.format(storage)

    logfile = os.path.join(recipe.buildout_dir, 'var', 'log', 'pack.log')
    return ' \\\n    && echo {datecmd} "{message}" >> {logfile}'.format(
        datecmd='`date +%Y-%m-%dT%H:%M:%S%z`',
        message=message,
        logfile=logfile)


def subpart_option(buildout, part, subpart, option, default=''):
    """Get an option for a filestorage subpart, falling back to the main part.
    """
    parts_to_check = ['%s_%s' % (part, subpart), part]
    val = default
    for p in parts_to_check:
        if p not in buildout:
            continue
        if option in buildout[p]:
            val = buildout[p][option]
            break

    return val % dict(fs_part_name=subpart)



from ftw.recipe.deployment.utils import chmod_executable
from zc.buildout.buildout import MissingOption
from zc.buildout import UserError
import os.path


def create_rc_scripts(recipe):
    """Create run-control scripts and return a list of all files that were
       created.
    """
    files = []

    startup_dir = recipe.options.get('startup-directory', None)
    shutdown_dir = recipe.options.get('shutdown-directory', None)
    rc_prefix = recipe.options.get('rc-prefix', 'rc-')
    rc_dir = recipe.options.get('rc-directory', None)
    rc_user = recipe.options.get('rc-user', 'zope')

    if startup_dir and not shutdown_dir:
        raise MissingOption("Option 'shutdown-directory' is required when "
                            "'startup-directory' is provided")
    if shutdown_dir and not startup_dir:
        raise MissingOption("Option 'startup-directory' is required when "
                            "'shutdown-directory' is provided")

    if startup_dir or shutdown_dir:
        # Try to create the startup directory
        if not os.path.isdir(startup_dir):
            try:
                os.makedirs(startup_dir)
            except OSError:
                pass
        # Try to create the shutdown directory
        if not os.path.isdir(shutdown_dir):
            try:
                os.makedirs(shutdown_dir)
            except OSError:
                pass

        if recipe.has_supervisor:
            # Create startup script
            script_path = os.path.join(startup_dir, recipe.buildout_name)
            script_file = open(script_path, 'w')
            script_file.write('#!/bin/sh\n%s/bin/supervisord\n' % recipe.buildout_dir)
            script_file.close()
            chmod_executable(script_path)
            files.append(script_path)
            # Create shutdown script
            script_path = os.path.join(shutdown_dir, recipe.buildout_name)
            script_file = open(script_path, 'w')
            script_file.write('#!/bin/sh\n%s/bin/supervisorctl shutdown\n' % recipe.buildout_dir)
            script_file.close()
            chmod_executable(script_path)
            files.append(script_path)
        else:
            raise UserError('supervisor section required with option startup_dir')

    else:
        if rc_dir is None:
            rc_dir = os.path.join(recipe.buildout_dir, 'bin')
        if not rc_dir:
            return files

        if recipe.has_supervisor:
            supervisorctl = '%s/bin/supervisorctl' % recipe.buildout_dir
            supervisord = '%s/bin/supervisord' % recipe.buildout_dir
            rc_filename = '%s/%ssupervisor' % (rc_dir, rc_prefix)
            rc_file = open(rc_filename, 'w')
            rc_file.write(SUPERVISOR_RC_TEMPLATE % dict(
                supervisorctl=supervisorctl, supervisord=supervisord, user=rc_user))
            rc_file.close()
            chmod_executable(rc_filename)
            files.append(rc_filename)
        else:
            for zope_part in recipe.zope_parts:
                zopectl = '%s/bin/%s' % (recipe.buildout_dir, zope_part)
                rc_filename = '%s/%s%s' % (rc_dir, rc_prefix, zope_part)
                rc_file = open(rc_filename, 'w')
                rc_file.write(ZOPE_RC_TEMPLATE % dict(zopectl=zopectl, user=rc_user))
                rc_file.close()
                chmod_executable(rc_filename)
                files.append(rc_filename)

            for zeo_part in recipe.zeo_parts:
                zeoctl = '%s/bin/%s' % (recipe.buildout_dir, zeo_part)
                rc_filename = '%s/%s%s' % (rc_dir, rc_prefix, zeo_part)
                rc_file = open(rc_filename, 'w')
                rc_file.write(ZEO_RC_TEMPLATE % dict(zeoctl=zeoctl, user=rc_user))
                rc_file.close()
                chmod_executable(rc_filename)
                files.append(rc_filename)

    return files


ZOPE_RC_TEMPLATE = """#!/bin/sh

# chkconfig: 345 90 10
# description: Starts Zope

START_SCRIPT="%(zopectl)s"

[ -f $START_SCRIPT ] || exit 1

# Source function library.
. /etc/init.d/functions

RETVAL=0

if [ $(whoami) != "root" ]; then
    echo "You must be root."
    exit 1
fi

case $1 in
    start|stop)
        su %(user)s -c "$START_SCRIPT $*" </dev/null
        RETVAL=$?
        if [ $RETVAL -eq 0 ]
        then
            echo_success
        else
            echo_failure
        fi
        echo
        ;;
    restart)
        ${0} stop
        sleep 1
        ${0} start
        ;;
    *)
        su %(user)s -c "$START_SCRIPT $*" </dev/null
        ;;
esac
"""

ZEO_RC_TEMPLATE = """#!/bin/sh

# chkconfig: 345 85 15
# description: Starts ZEO server

START_SCRIPT="%(zeoctl)s"

[ -f $START_SCRIPT ] || exit 1

# Source function library.
. /etc/init.d/functions

RETVAL=0

if [ $(whoami) != "root" ]; then
    echo "You must be root."
    exit 1
fi

case $1 in
    start|stop)
        su %(user)s -c "$START_SCRIPT $*" </dev/null
        RETVAL=$?
        if [ $RETVAL -eq 0 ]
        then
            echo_success
        else
            echo_failure
        fi
        echo
        ;;
    restart)
        ${0} stop
        sleep 1
        ${0} start
        ;;
    *)
        su %(user)s -c "$START_SCRIPT $*" </dev/null
        ;;
esac
"""

SUPERVISOR_RC_TEMPLATE = """#!/bin/sh

# chkconfig: 345 90 10
# description: Starts supervisor

SUPERVISORCTL="%(supervisorctl)s"
SUPERVISORD="%(supervisord)s"

[ -f $SUPERVISORD ] || exit 1
[ -f $SUPERVISORCTL ] || exit 1

# Source function library.
. /etc/init.d/functions

RETVAL=0

if [ $(whoami) != "root" ]; then
    echo "You must be root."
    exit 1
fi

start() {
    echo -n "Starting supervisor: "
    su %(user)s -c "$SUPERVISORD"
    RETVAL=$?
    if [ $RETVAL -eq 0 ]; then
        echo_success
    else
        echo_failure
    fi
    return $RETVAL
}

stop() {
    echo -n "Stopping supervisor: "
    su %(user)s -c "$SUPERVISORCTL shutdown"
    RETVAL=$?
    if [ $RETVAL -eq 0 ]; then
        echo_success
    else
        echo_failure
    fi
    return $RETVAL
}

status() {
    su %(user)s -c "$SUPERVISORCTL status"
}

case "$1" in
    start)
        start
        ;;

    stop)
        stop
        ;;

    restart)
        stop
        start
        ;;

    status)
        status
        ;;
esac

exit $REVAL
"""



# -*- coding: utf-8 -*-
import os
from setuptools import setup, find_packages

version = '1.4.3'
tests_require = ['zc.buildout [test]']

setup(name='ftw.recipe.deployment',
      version=version,
      description="A zc.buildout recipe for Plone deployments which configures"
                  " various unix system services.",
      long_description=open("README.rst").read()
                       + "\n" +
                       open("HISTORY.txt").read()
                       + "\n" +
                       open(os.path.join("ftw", "recipe", "deployment",
                                         "README.txt")).read(),

      classifiers=[
        'Framework :: Buildout :: Recipe',
        'Intended Audience :: Developers',
        'License :: OSI Approved :: Zope Public License',
        'Programming Language :: Python',
        'Topic :: Software Development :: Build Tools',
        ],

      keywords='',
      author='4teamwork AG',
      author_email='mailto:info@4teamwork.ch',
      url='https://github.com/4teamwork/ftw.recipe.deployment',
      license='GPL2',

      packages=find_packages(exclude=['ez_setup']),
      namespace_packages=['ftw', 'ftw.recipe'],
      include_package_data=True,
      zip_safe=False,
      install_requires=[
        'setuptools',
        'zc.buildout',
        ],

      tests_require=tests_require,
      extras_require=dict(tests=tests_require),

      test_suite='ftw.recipe.deployment.tests.test_docs.test_suite',
      entry_points = {
        'zc.buildout': ['default = ftw.recipe.deployment:Recipe']},
      )



import os
import os.path

PROSPECTOR_TEMPLATE = """\
- type: log
  fields:
    event_type: %s
    deployment: %s
  fields_under_root: true
  json.keys_under_root: true
  json.add_error_key: true
  paths:
    - %s\
"""


def create_filebeat_conf(recipe):
    """Create a filebeat prospectors config file and return the filename of
    the created file.
    """
    default_etc_dir = os.path.join(recipe.buildout_dir, 'etc')
    etc_dir = recipe.options.get('etc-directory', default_etc_dir)

    filebeat_dir = os.path.join(etc_dir, 'filebeat.d')

    # Try to create the filebeat config directory
    if not os.path.isdir(filebeat_dir):
        try:
            os.makedirs(filebeat_dir)
        except OSError:
            return None

    # Create contentstats prospector (one per deployment)
    contentstats_log = os.path.join(
        recipe.buildout_dir, 'var/log/contentstats-json.log')
    contentstats_prospector = PROSPECTOR_TEMPLATE % (
        'contentstats', recipe.buildout_name, contentstats_log)

    # Add configuration for ftw.structlog logs (one per instance)
    prospectors = [contentstats_prospector]
    for zope_part in recipe.zope_parts:
        event_log = '%s/var/log/%s.log' % (recipe.buildout_dir, zope_part)
        event_log = recipe.buildout[zope_part].get('event-log', event_log)
        # ftw.structlog's log path is based on eventlog path
        structlog_json_log = event_log.replace('.log', '-json.log')

        structlog_prospector = PROSPECTOR_TEMPLATE % (
            'structlog', recipe.buildout_name, structlog_json_log)
        prospectors.append(structlog_prospector)

    # Create the filebeat config file
    file_name = '%s.yml' % os.path.join(filebeat_dir, recipe.buildout_name)
    filebeat_file = open(file_name, 'w')
    filebeat_file.write('\n'.join(prospectors))
    filebeat_file.close()
    return file_name



# -*- coding: utf-8 -*-
"""Recipe deployment"""
import os.path
from ftw.recipe.deployment.filebeat import create_filebeat_conf
from ftw.recipe.deployment.logrotate import create_logrotate_conf
from ftw.recipe.deployment.pack import create_pack_script
from ftw.recipe.deployment.rc import create_rc_scripts
import logging


recipe_logger = logging.getLogger(__name__)


class Recipe(object):
    """zc.buildout recipe"""

    def __init__(self, buildout, name, options):
        self.buildout, self.name, self.options = buildout, name, options
        part_names = [p.strip() for p in self.buildout['buildout']['parts']
                      .split()]

        # Figure out zope instance parts
        self.zope_parts = options.get('zopes', '').split()
        if len(self.zope_parts) == 0:
            for part_name in part_names:
                part = self.buildout[part_name]
                if part.get('recipe', None) == 'plone.recipe.zope2instance':
                    self.zope_parts.append(part_name)

        if not all_unique(self.zope_parts):
            msg = '\n'.join((
                '',
                '##########################################################',
                '# Duplicate zope parts found!',
                '# %r',
                '# This could be caused by extending from multiple zeoclients/',
                '# .cfg files instead of just one, and should be avoided.',
                '# To prevent generating invalid logrotate configs, duplicate',
                '# parts will be skipped, but you should still fix this issue.',
                '##########################################################',
                '',
            )) % self.zope_parts

            recipe_logger.warn(msg)
            self.zope_parts = sorted(list(set(self.zope_parts)))

        # Figure out zeo server parts
        self.zeo_parts = options.get('zeos', '').split()
        if len(self.zeo_parts) == 0:
            for part_name in part_names:
                part = self.buildout[part_name]
                if part.get('recipe', None) in ['plone.recipe.zope2zeoserver',
                                                'plone.recipe.zeoserver']:
                    self.zeo_parts.append(part_name)

        # Figure out if supervisor is beeing used
        self.has_supervisor = options.get('has-supervisor', None)
        if self.has_supervisor is None:
            self.has_supervisor = False
            for part_name in part_names:
                part = self.buildout[part_name]
                if part.get('recipe', None) == 'collective.recipe.supervisor':
                    self.has_supervisor = True
                    break
        else:
            if self.has_supervisor.lower() == 'false':
                self.has_supervisor = False
            else:
                self.has_supervisor = True

        # Determine if we should create filebeat config
        self.create_filebeat_config = options.get('create-filebeat-config', '')
        if self.create_filebeat_config.lower() == 'false':
            self.create_filebeat_config = False
        else:
            self.create_filebeat_config = True

        # Figure out filestorage parts
        self.filestorage_parts = options.get('filestorage', '').split()
        if len(self.filestorage_parts) == 0:
            for part_name in part_names:
                part = self.buildout[part_name]
                if part.get('recipe', None) == 'collective.recipe.filestorage':
                    self.filestorage_parts.append(part_name)

        self.buildout_dir = self.buildout['buildout']['directory']
        self.buildout_name = os.path.basename(self.buildout_dir)

    def install(self):
        """Installer"""
        files = []

        # Create logrotate configuration
        logrotate_conf = create_logrotate_conf(self)
        if logrotate_conf:
            files.append(logrotate_conf)

        # Create filebeat configuration
        if self.create_filebeat_config:
            filebeat_conf = create_filebeat_conf(self)
            if filebeat_conf:
                files.append(filebeat_conf)

        rc_scripts = create_rc_scripts(self)
        files.extend(rc_scripts)

        files.extend(create_pack_script(self))

        # Return files that were created by the recipe. The buildout
        # will remove all returned files upon reinstall.
        return files

    def update(self):
        """Updater"""
        return self.install()


def all_unique(sequence):
    return len(set(sequence)) == len(sequence)



import os
import os.path
from StringIO import StringIO


def create_logrotate_conf(recipe):
    """Create a logrotate configuration file and return the filename of the
       created file.
    """
    logrotate_dir = recipe.options.get('logrotate-directory', None)
    if logrotate_dir is None:
        logrotate_dir = os.path.join(recipe.buildout_dir,
                                     'etc', 'logrotate.d')
    if not logrotate_dir:
        return None

    # Try to create the logrotate directory
    if not os.path.isdir(logrotate_dir):
        try:
            os.makedirs(logrotate_dir)
        except OSError:
            return None

    logrotate_options = recipe.options.get('logrotate-options',
                                           '').splitlines()
    logrotate_options = [opt for opt in logrotate_options if opt]
    logrotate_conf = StringIO()

    # Add configuration for zope parts
    for zope_part in recipe.zope_parts:
        logs = []
        event_log = '%s/var/log/%s.log' % (recipe.buildout_dir, zope_part)
        event_log = recipe.buildout[zope_part].get('event-log', event_log)
        if event_log:
            logs.append(event_log)
        z2_log = '%s/var/log/%s-Z2.log' % (recipe.buildout_dir, zope_part)
        z2_log = recipe.buildout[zope_part].get('z2-log', z2_log)
        if z2_log:
            logs.append(z2_log)
        if not logs:
            continue

        logrotate_conf.write('\n'.join(logs))
        logrotate_conf.write(' {\n')

        # logrotate options
        for opt in logrotate_options:
            logrotate_conf.write('    %s\n' % opt)

        # add reopening of logs using postrotate script
        if 'postrotate' not in logrotate_options:
            logrotate_conf.write('    sharedscripts\n')
            logrotate_conf.write('    postrotate\n')
            logrotate_conf.write('        /bin/kill -SIGUSR2 `cat %s/var/'
                                 '%s.pid 2>/dev/null` >/dev/null 2>&1 || true'
                                 '\n' % (
                                 recipe.buildout_dir, zope_part))
            logrotate_conf.write('    endscript\n')

        logrotate_conf.write('}\n')

    # Add configuration for zeo part
    for zeo_part in recipe.zeo_parts:
        zeo_log = '%s/var/log/%s.log' % (recipe.buildout_dir, zeo_part)
        zeo_log = recipe.buildout[zeo_part].get('zeo-log', zeo_log)
        if not zeo_log:
            continue

        logrotate_conf.write('%s {\n' % zeo_log)

        # logrotate options
        for opt in logrotate_options:
            logrotate_conf.write('    %s\n' % opt)

        # logreopen is broken in ZODB <= 3.9.6
        # use copytruncate instead
        if 'nocopytruncate' not in logrotate_options:
            logrotate_conf.write('    copytruncate\n')

        logrotate_conf.write('}\n')

    # Add configuration for ftw.structlog logs
    for zope_part in recipe.zope_parts:
        event_log = '%s/var/log/%s.log' % (recipe.buildout_dir, zope_part)
        event_log = recipe.buildout[zope_part].get('event-log', event_log)
        # ftw.structlog's log path is based on eventlog path
        structlog_json_log = event_log.replace('.log', '-json.log')

        if event_log:
            logrotate_conf.write(structlog_json_log)
            logrotate_conf.write(' {\n')

            # logrotate options
            for opt in logrotate_options:
                if opt == 'nocopytruncate':
                    # We always want copytruncate for ftw.structlog logfiles
                    continue
                logrotate_conf.write('    %s\n' % opt)

            if 'copytruncate' not in logrotate_options:
                # Always use 'copytruncate' for ftw.structlog logfiles. We need
                # this because ftw.structlog doesn't have a signal handler to
                # reopen logfiles after rotation
                logrotate_conf.write('    copytruncate\n')

            # Ensure 'missingok' is always set, since ftw.structlog logfiles
            # may not exist for many deplyoments, but we always generate the
            # logrotation config for them unconditionally
            if 'missingok' not in logrotate_options:
                logrotate_conf.write('    missingok\n')

            logrotate_conf.write('}\n')

    # Create the logrotate file
    file_name = os.path.join(logrotate_dir, recipe.buildout_name)
    logrotate_file = open(file_name, 'w')
    logrotate_file.write(logrotate_conf.getvalue())
    logrotate_file.close()
    return file_name
